# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):

        # Changing field 'Tag.tag'
        db.alter_column(u'geocode_tag', 'tag', self.gf('django.db.models.fields.CharField')(unique=True, max_length=4096))

        # Changing field 'GeoData.tags'
        db.alter_column(u'geocode_geodata', 'tags', self.gf('django.db.models.fields.CharField')(max_length=4096))

        # Changing field 'GeoAddress.cluster'
        db.alter_column(u'geocode_geoaddress', 'cluster', self.gf('django.db.models.fields.CharField')(max_length=4096, null=True))

    def backwards(self, orm):

        # Changing field 'Tag.tag'
        db.alter_column(u'geocode_tag', 'tag', self.gf('django.db.models.fields.CharField')(max_length=50, unique=True))

        # Changing field 'GeoData.tags'
        db.alter_column(u'geocode_geodata', 'tags', self.gf('django.db.models.fields.CharField')(max_length=255))

        # Changing field 'GeoAddress.cluster'
        db.alter_column(u'geocode_geoaddress', 'cluster', self.gf('django.db.models.fields.CharField')(max_length=50, null=True))

    models = {
        u'geocode.geoaddress': {
            'Meta': {'ordering': "('-created',)", 'object_name': 'GeoAddress'},
            'address': ('django.db.models.fields.TextField', [], {}),
            'cluster': ('django.db.models.fields.CharField', [], {'max_length': '4096', 'null': 'True'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'uuid': ('uuidfield.fields.UUIDField', [], {'unique': 'True', 'max_length': '32'})
        },
        u'geocode.geodata': {
            'Meta': {'ordering': "('-created',)", 'object_name': 'GeoData'},
            'address': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['geocode.GeoAddress']"}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'point': ('django.contrib.gis.db.models.fields.PointField', [], {}),
            'tags': ('django.db.models.fields.CharField', [], {'max_length': '4096'}),
            'weight': ('django.db.models.fields.PositiveIntegerField', [], {'default': '1'})
        },
        u'geocode.tag': {
            'Meta': {'object_name': 'Tag'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'tag': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '4096'})
        }
    }

    complete_apps = ['geocode']


# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'GeoAddress'
        db.create_table(u'geocode_geoaddress', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('uuid', self.gf('uuidfield.fields.UUIDField')(unique=True, max_length=32)),
            ('cluster', self.gf('django.db.models.fields.CharField')(max_length=50, null=True)),
            ('address', self.gf('django.db.models.fields.TextField')()),
        ))
        db.send_create_signal(u'geocode', ['GeoAddress'])

        # Adding model 'GeoData'
        db.create_table(u'geocode_geodata', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('address', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['geocode.GeoAddress'])),
            ('point', self.gf('django.contrib.gis.db.models.fields.PointField')()),
            ('tags', self.gf('django.db.models.fields.CharField')(max_length=255)),
        ))
        db.send_create_signal(u'geocode', ['GeoData'])

        # Adding model 'Tag'
        db.create_table(u'geocode_tag', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('tag', self.gf('django.db.models.fields.CharField')(unique=True, max_length=50)),
        ))
        db.send_create_signal(u'geocode', ['Tag'])


    def backwards(self, orm):
        # Deleting model 'GeoAddress'
        db.delete_table(u'geocode_geoaddress')

        # Deleting model 'GeoData'
        db.delete_table(u'geocode_geodata')

        # Deleting model 'Tag'
        db.delete_table(u'geocode_tag')


    models = {
        u'geocode.geoaddress': {
            'Meta': {'ordering': "('-created',)", 'object_name': 'GeoAddress'},
            'address': ('django.db.models.fields.TextField', [], {}),
            'cluster': ('django.db.models.fields.CharField', [], {'max_length': '50', 'null': 'True'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'uuid': ('uuidfield.fields.UUIDField', [], {'unique': 'True', 'max_length': '32'})
        },
        u'geocode.geodata': {
            'Meta': {'ordering': "('-created',)", 'object_name': 'GeoData'},
            'address': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['geocode.GeoAddress']"}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'point': ('django.contrib.gis.db.models.fields.PointField', [], {}),
            'tags': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        u'geocode.tag': {
            'Meta': {'object_name': 'Tag'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'tag': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '50'})
        }
    }

    complete_apps = ['geocode']


from django.conf import settings

MIN_GEOCODES = getattr(settings, 'GEOCODE_MIN_GEOCODES', 2)
DATAPOINTS = getattr(settings, 'GEOCODE_DATAPOINTS', 50)

GEOCODERS = getattr(settings, 'GEOCODERS', (
    # Good open GeoCoders
    ('geopy.geocoders.GeocodeFarm', {}),
    ('geopy.geocoders.GoogleV3', {}),

    # Un-reliable open GeoCoders
    # ('geopy.geocoders.ArcGIS', {}),
    # ('geopy.geocoders.DataBC', {}),
    # ('geopy.geocoders.GeocoderDotUS', {}),
    # ('geopy.geocoders.NaviData', {}),
    # ('geopy.geocoders.Nominatim', {}),
    # ('geopy.geocoders.OpenMapQuest', {}),
    # ('geopy.geocoders.Yandex', {}),

    # Closed GeoCoders
    # ('geopy.geocoders.Baidu', {'api_key': ''}),
    # ('geopy.geocoders.Bing', {'api_key': ''}),
    # ('geopy.geocoders.GeoNames', {'username': ''}),
    # ('geopy.geocoders.IGNFrance', {'api_key': '',
    #                                'username': '',
    #                                'password': ''}),
    # ('geopy.geocoders.LiveAddress', {'auth_id': '', 'auth_token': ''}),
    # ('geopy.geocoders.OpenCage', {'api_key': ''}),
    # ('geopy.geocoders.YahooPlaceFinder', {'consumer_key': '',
    #                                       'consumer_secret': ''}),
    # ('geopy.geocoders.What3Words', {'api_key': ''}),
))



from __future__ import absolute_import

from celery import Celery
from django.conf import settings

app = Celery('core')
app.config_from_object('django.conf:settings')
app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)



from django.test import TestCase
from geopy.distance import vincenty
from geopy.location import Location
from geopy.point import Point
from mock import patch, ANY


def compare(test_value, other, accuratcy):
    if not type(test_value) == type(other):
        return False

    # Coordinates are equal if within x meters.
    distance_meters = vincenty(test_value, other).m
    if distance_meters > accuratcy:
        msg = 'Distance {0} greater then {1} meters.'.format(
            distance_meters, accuratcy)
        raise AssertionError(msg)

    return True


class Matcher(object):
    def __init__(self, compare, test_value, accuratcy):
        self.compare = compare
        self.test_value = test_value
        self.accuratcy = accuratcy

    def __eq__(self, other):
        return self.compare(self.test_value, other, self.accuratcy)


class GeoCodeTestCase(TestCase):

    def get_patch(self, name):
        patcher = patch(name)
        thing = patcher.start()
        self.addCleanup(patcher.stop)
        return thing

    def setUp(self):
        self.geocoders = []
        good_geocoders = [
            'geopy.geocoders.GeocodeFarm.geocode',
        ]

        for geocoder in good_geocoders:
            geo = self.get_patch(geocoder)
            geo.return_value = Location(point=Point([48.8582653, 2.2944946]))
            self.geocoders.append(geo)

        bad_geocoders = [
            'geopy.geocoders.GoogleV3.geocode',
        ]

        for geocoder in bad_geocoders:
            geo = self.get_patch(geocoder)
            geo.return_value = Location(point=Point([48.8581653, 2.2944946]))
            self.geocoders.append(geo)

    @patch('geocode.geocode_update.send')
    def test_initial_geocode(self, geocode_update):
        """
        Test GEO coding of initial address.
        """
        from geocode import geocode_address
        from uuid import uuid4

        point_uuid = uuid4()
        geocode_address(point_uuid, 'Eiffel Tower, Paris, France')

        match_coordinates = Matcher(compare, [48.8582653, 2.2944946], 6)
        geocode_update.assert_called_once_with(sender=ANY,
                                               cluster=None,
                                               point_uuid=point_uuid,
                                               coordinates=match_coordinates)

        self.assertEqual(len(self.geocoders), 2)
        for geocoder in self.geocoders:
            geocoder.assert_called_once_with('Eiffel Tower, Paris, France')

    @patch('geocode.geocode_update.send')
    def test_add_coordinates(self, geocode_update):
        """
        Test adding extra details.
        """
        from geocode import geocode_address, add_coordinates
        from uuid import uuid4

        point_uuid = uuid4()
        geocode_address(point_uuid, 'Eiffel Tower, Paris, France',
                        cluster='test')
        geocode_update.reset_mock()

        add_coordinates(point_uuid, [48.8582653, 2.2944946], tags=['ios', ])

        match_coordinates = Matcher(compare, [48.8582653, 2.2944946], 4)
        geocode_update.assert_called_once_with(sender=ANY,
                                               cluster='test',
                                               point_uuid=point_uuid,
                                               coordinates=match_coordinates)

    @patch('geocode.geocode_update.send')
    def test_regression_disparate_coordinates(self, geocode_update):
        """
        Regression test for when a mean center can not be found. This happens
        if all coordinates are far apart.
        """
        from geocode.models import GeoAddress
        from uuid import uuid4

        point_uuid = uuid4()
        address = GeoAddress.objects.create(uuid=point_uuid, cluster='reg')
        address.add_coordinate([48.8582653, 2.2944946], ['ios', ])
        address.add_coordinate([48.8582653, 1.2944946], ['ios', ])
        address.add_coordinate([48.8582653, 0.2944946], ['ios', ])
        address.send_geocode_update()

        match_coordinates = Matcher(compare, [48.8582653, 1.2944946], 3)
        geocode_update.assert_called_once_with(sender=ANY,
                                               cluster='reg',
                                               point_uuid=point_uuid,
                                               coordinates=match_coordinates)



from django.core.urlresolvers import reverse
from django.utils.html import mark_safe
from geocode.models import GeoAddress


def geocode_address(point_uuid, address, cluster=None):
    """
    Geo-codes an address. The result is returned in a signal.

    point_uuid: Unique identifier for the address in the system.
    cluster: A logical bucket to put the address in. This allows listeners to
        the signal to identify messages they are interested in.
    address: Address to be GEO coded.
    """
    address, created = GeoAddress.objects.get_or_create(uuid=point_uuid,
                                                        cluster=cluster,
                                                        address=address)
    address.run_geocode()
    return address


def add_coordinates(point_uuid, coordinates, tags=None, weight=None):
    """
    Adds coordinates to an address. A signal will be generated with updated
    coordinates.

    point_uuid: Unique identifier for the address in the system.
    coordinates: Coordinates to be added to the address data.
    """
    address = GeoAddress.objects.get(uuid=point_uuid)
    address.add_coordinate(coordinates, tags, weight)
    address.send_geocode_update()
    return address


def delete_coordinates(point_uuid, tags=None, geocoders=False):
    """
    Delete coordinates from an address. A signal will be generated with updated
    coordinates.

    point_uuid: Unique identifier for the address in the system.
    tag: Only delete coordinates with this tag.
    """
    address = GeoAddress.objects.get(uuid=point_uuid)
    address.delete_coordinate(tags=tags, geocoders=geocoders)
    address.send_geocode_update()
    return address


class GeocodeAdmin(object):

    def geocode_data(self, obj):
        """
        Allows adding link to geocode data from an objects admin change view.
        """
        geocode = GeoAddress.objects.get(uuid=obj.point_uuid)
        url = reverse('admin:geocode_geoaddress_change', args=[geocode.id])
        return mark_safe('<a href="{0}">{1}</a>'.format(url, obj.point_uuid))



from django.contrib.gis.db import models
from django.contrib.gis.db.models import Q
from django.contrib.gis.geos import Point
from django.utils.translation import ugettext_lazy as _

try:
    from django.db.models import UUIDField
    # Django 1.8 now includes a UUIDField (database representation differs)
except ImportError:
    from uuidfield import UUIDField

from geocode import conf
from geocode.utils import calculate_center


class GeoAddress(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    uuid = UUIDField(unique=True)
    cluster = models.CharField(_('Cluster'), max_length=4096, null=True)
    address = models.TextField()

    class Meta:
        verbose_name = _('Geo address')
        verbose_name_plural = _('Geo addresses')
        ordering = ('-created', )

    def __unicode__(self):
        return self.address

    @property
    def coordinates(self):
        """
        Return the estimated location based on the GEO-coding results.
        Currently we are simply ignoring anything with a distance from the mean
        center that is greater then the 80 percentile of all distances from the
        mean center. Falling back to the mean center if all point are outside
        the 80 percentile.
        """
        points = []
        for point in self.geodata_set.all()[:conf.DATAPOINTS]:
            for i in range(0, point.weight):
                points.append(point.point)
        center = calculate_center(points)
        return [center.x, center.y]

    @property
    def point(self):
        return Point(self.coordinates[::-1])

    def run_geocode(self):
        from geocode.tasks import geocode_address_task
        geocode_address_task.delay(self.uuid)

    def send_geocode_update(self):
        from geocode.signals import geocode_update
        geocode_update.send(sender=GeoAddress,
                            point_uuid=self.uuid,
                            cluster=self.cluster,
                            coordinates=self.coordinates)

    def add_coordinate(self, coordinates, tags=None, weight=None):
        if tags is None:
            tags = ['not-set', ]

        if weight is None:
            weight = 1

        tags = [tag.lower() for tag in tags]
        for tag in tags:
            Tag.objects.get_or_create(tag=tag)
        tags = ','.join(tags)

        self.geodata_set.create(point=Point(*coordinates), tags=tags,
                                weight=weight)

    def delete_coordinate(self, tags=None, geocoders=False):
        if tags is None:
            tags = []

        if geocoders:
            for import_path, __ in conf.GEOCODERS:
                tags.append(import_path.split('.')[-1])

        q_objects = Q()
        for tag in tags:
            q_objects |= Q(tags__icontains=tag)

        self.geodata_set.filter(q_objects).delete()


class GeoData(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    address = models.ForeignKey(GeoAddress)
    point = models.PointField(srid=4326)
    tags = models.CharField(max_length=4096)
    weight = models.PositiveIntegerField(default=1)

    class Meta:
        verbose_name = _('Geo data')
        verbose_name_plural = _('Geo data')
        ordering = ('-created', )

    def __unicode__(self):
        return '{0.x}, {0.y}'.format(self.point)


class Tag(models.Model):
    tag = models.CharField(_('Tag'), max_length=4096, unique=True)

    class Meta:
        verbose_name = _('Tag')
        verbose_name_plural = _('Tags')

    def __unicode__(self):
        return self.tag



import math
from geopy.distance import vincenty
from django.contrib.gis.geos import Point, MultiPoint


def percentile(data_list, value):
    data_list = list(map(float, data_list))
    count = float(len(data_list))
    strict = len([i for i in data_list if i < value]) / count * 100
    weak = len([i for i in data_list if i <= value]) / count * 100
    return mean([strict, weak])


def mean(data_list):
    length = len(data_list)
    if length == 0:
        return 0
    data_list = list(map(float, data_list))
    return sum(data_list) / length


def mean_center(points):
    return MultiPoint(points).centroid


def standard_deviation(data_list):
    data_list = list(map(float, data_list))
    mean_value = mean(data_list)
    deviations = [i - mean_value for i in data_list]
    deviations_squared = [math.pow(i, 2) for i in deviations]
    mean_deviation = mean(deviations_squared)
    return math.sqrt(mean_deviation)


def mean_distances(points):
    mean = mean_center(points)
    return mean, [vincenty(p, mean).meters for p in points]


def calculate_center(points, depth=0):
    """
    Finds the mean center of the provided points. Returns when it finds a
    cluster of points who's mean distance from the mean center is less
    then 10 meters or when the max depth (20) is reached. On each iteration it
    removes the furthest 5 percentile from the center.
    """

    if len(points) == 0:
        return Point(0, 0)

    center, distances = mean_distances(points)
    if mean(distances) <= 10 or depth > 20:
        return center

    percentiles = [percentile(distances, d) for d in distances]
    max_percentile = max(percentiles) * 0.95

    center_points = []
    center_distences = []
    for i, point_percentile in enumerate(percentiles):
        if point_percentile <= max_percentile:
            center_points.append(points[i])
            center_distences.append(distances[i])

    if len(center_points):
        return calculate_center(center_points, depth=depth + 1)
    else:
        return center



from geopy.exc import GeopyError
from celery import shared_task

from geocode import conf
from geocode.models import GeoAddress

try:
    # Django 1.7 and onwards.
    from django.utils.module_loading import import_string
except ImportError:
    # Older version of Django.
    from django.utils.module_loading import import_by_path as import_string


geocoders = []
for import_path, kwargs in conf.GEOCODERS:
    Geocoder = import_string(import_path)
    geocoders.append((Geocoder(**kwargs), import_path.split('.')[-1]))


@shared_task(bind=True, default_retry_delay=60*3, max_retries=20,
             ignore_result=True)
def geocode_address_task(self, point_uuid):
    address = GeoAddress.objects.get(uuid=point_uuid)

    success_count = 0
    for geocoder, tag in geocoders:
        try:
            location = geocoder.geocode(address.address)
        except (GeopyError, IndexError):
            pass
        else:
            if location:
                coordinates = [location.latitude, location.longitude]
                address.add_coordinate(coordinates, tags=[tag, ])
                success_count += 1

    if success_count < min(conf.MIN_GEOCODES, len(geocoders)):
        self.retry()

    address.send_geocode_update()



from django.contrib.gis import admin
from geocode.models import GeoAddress, GeoData


class GeoDataInline(admin.TabularInline):
    model = GeoData
    extra = 0
    fields = ('created', 'address', 'tags', 'point', 'weight')
    readonly_fields = ('created', 'address', 'tags', 'point', 'weight')

    def has_add_permission(self, *args):
        return False


class GeoAddressAdmin(admin.OSMGeoAdmin):
    inlines = [GeoDataInline, ]
    list_display = ('created', 'uuid', 'cluster', 'address')
    search_fields = ('address', )
    fields = ('created', 'uuid', 'cluster', 'address', 'coordinates')
    readonly_fields = ('created', 'uuid', 'cluster', 'address', 'coordinates')

    def has_add_permission(self, *args):
        return False

    def has_delete_permission(self, *args):
        return False

admin.site.register(GeoAddress, GeoAddressAdmin)



#!/usr/bin/env python

import os
import sys

# fix sys path so we don't need to setup PYTHONPATH
sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
os.environ['DJANGO_SETTINGS_MODULE'] = 'geocode.runtests.settings'


import django
from django.conf import settings
from django.test.utils import get_runner


def main():
    try:
        django.setup()
    except AttributeError:
        pass
    TestRunner = get_runner(settings)
    test_runner = TestRunner()
    failures = test_runner.run_tests(['geocode.tests'])
    sys.exit(failures)


if __name__ == '__main__':
    main()
